# -*- coding: utf-8 -*-
"""
Created on Fri Oct 15 13:16:43 2021

@author: Nebojsa.Vojinovic
"""
# This allows you to connect to the Workspace you've previously deployed in Azure.
# Be sure to fill in the settings below which can be retrieved by running 'az quantum workspace show' in the terminal.
from azure.quantum import Workspace

# Copy the settings for your workspace below
workspace = Workspace (
  subscription_id = "",
  resource_group = "",
  name = "",
  location = ""
)



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 10 17:12:10 2021

@author: Nebojsa.Vojinovic
"""

from dimod import ExactSolver
from dwave.system import DWaveSampler, EmbeddingComposite
from dwave.cloud.exceptions import SolverNotFoundError
from dwave.system.samplers import LeapHybridSampler


class Solvers:
  def __init__(self, num_reads, chain_strength):
    try:
        qpu_advantage = DWaveSampler(solver={'topology__type': 'pegasus', 'qpu': True})
#        qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera', 'qpu': True})
        exact = ExactSolver()
        hybrid = LeapHybridSampler()
        
        self.mSolvers = {'Advantage': EmbeddingComposite(qpu_advantage), 
#                         'DW-2000Q': EmbeddingComposite(qpu_2000q), 
                         'Exact': exact,
                         'Hybrid': hybrid
                         }
    
#        print(">>>> Connected to Advantage {} and 2000Q {}.".format(qpu_advantage.solver.id, qpu_2000q.solver.id))
    except SolverNotFoundError:
        print(">>>> Currently a pair of solvers are unavailable for sections comparing QPU technologies. Try those examples later.")
    
    
    # Return num_reads solutions (responses are in the D-Wave's graph of indexed qubits)
    # strength 5 for R 16, 15, 6, 12 works for both, pegasus and chimera
    self.mKwargs = {'num_reads': num_reads,
                    'answer_mode': 'histogram',
                    'chain_strength': chain_strength   
                    }


  def solve(self, solverType, qubo):
    solver = self.mSolvers[solverType]
    if solverType == 'Advantage':
        sampleset = solver.sample_qubo(qubo, **self.mKwargs)
    else:
        sampleset = solver.sample_qubo(qubo)
    samples = [dict(sample) for sample in sampleset.lowest().samples()]
    return samples



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 15 13:16:43 2021

@author: Nebojsa.Vojinovic
"""
from qiskit import IBMQ

from qiskit.aqua import aqua_globals, QuantumInstance
from qiskit.aqua.algorithms import QAOA, NumPyMinimumEigensolver
from qiskit.optimization.algorithms import MinimumEigenOptimizer, RecursiveMinimumEigenOptimizer
from qiskit.optimization import QuadraticProgram

from dann5.d5o2 import Qanalyzer

import numpy as np

class QuantumRequest:
  provider = None
  
  def __init__(self, assignmnet, timeout=100):
    if IBMQ.active_account() == None:
        self.provider = IBMQ.load_account()
    else:
        self.provider = IBMQ.get_provider()
    self.assign = assignmnet
    self.tm = timeout
    self.problem = None
    self.result = None
    self.nodes = None
    self.branches = None
    

  def reset(self):
    self.problem = None
    self.result = None
    self.nodes = None
    self.branches = None
    self.assignment.reset()
    
    
  def execute(self):
    qubo = self.assign.qubo()
    self.qubo2problem(qubo)
    
    # Instantiate a solver to solve the problem.
    quantum_instance = QuantumInstance(self.provider.get_backend('ibmq_qasm_simulator'),
                                       seed_simulator=aqua_globals.random_seed,
                                       seed_transpiler=aqua_globals.random_seed)
    qaoa_mes = QAOA(quantum_instance=quantum_instance, initial_point=[0., 0.])
    qaoa = MinimumEigenOptimizer(qaoa_mes)   # using QAOA
    self.result = qaoa.solve(self.problem)
    #print(f'\nResult:\n{self.result}\n')
    self.solve()

    
  def qubo2problem(self, qubo) -> QuadraticProgram:
    analyzer = Qanalyzer(qubo)
    
    self.nodes = analyzer.nodes()
    energies = []
    
    self.problem = QuadraticProgram()
    for (node, energy) in self.nodes:
        self.problem.binary_var(node)
        energies.append(energy)
    
    self.branches = dict(analyzer.branches())
    self.problem.minimize(linear= np.array(energies), quadratic=self.branches)
    #print(self.problem.export_as_lp_string())
    return self.problem
    
        
  def solve(self):
    # Print a summary of the result
    d5osamples = []
    for sample in self.result.samples:
        if self.result.fval == sample.fval:
            d5osample = {}
            for qbit in range(len(sample.x)):
                value = int(sample.x[qbit])
                name = self.nodes[int(qbit)]
                d5osample[name[0]] = value
            d5osamples.append(d5osample)
    #print(f'The samples are: {d5osamples}')
    self.assign.add(d5osamples)



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 22 13:43:58 2021

@author: Nebojsa.Vojinovic

https://pypi.org/manage/account/
Username: voya-voja
Token: pypi-AgEIcHlwaS5vcmcCJDcwODM5ZTU4LTRjZjAtNGIyYS05YTllLTMxMWU0OGI5Njg0MQACJXsicGVybWlzc2lvbnMiOiAidXNlciIsICJ2ZXJzaW9uIjogMX0AAAYgJfMHimYQb1kn1OCHZtPQ34NtGc3jIod8pIWFH7-uiaU

To use this API token:

Set your username to __token__
Set your password to the token value, including the pypi- prefix

For example, if you are using Twine to upload your projects to PyPI, set up your $HOME/.pypirc file like this:

[pypi]
  username = __token__
  password = pypi-AgEIcHlwaS5vcmcCJDcwODM5ZTU4LTRjZjAtNGIyYS05YTllLTMxMWU0OGI5Njg0MQACJXsicGVybWlzc2lvbnMiOiAidXNlciIsICJ2ZXJzaW9uIjogMX0AAAYgJfMHimYQb1kn1OCHZtPQ34NtGc3jIod8pIWFH7-uiaU

"""
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='dann5',    # This is the name of your PyPI-package.
    version='2.0.7',                          # Update the version number for new releases
    author="Voya",
    author_email="nebojsa.vojinovic@rogers.com",
    description="A Library of Quantum computing types",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/voya-voja/d5o2",
    license='GPLv3',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator',
    ],
    keywords='quantum coding development types',
    project_urls={
        "Bug Tracker": "https://github.com/voya-voja/d5o2/issues",
    },
    packages=['dann5'],
#    packages=find_packages(include=['dann5', 'dann5.*']),
#    package_dir={'dann5': 'dann5'},
#    include_package_data = True,
#    package_data={'dann5': ['x64/Release/d5o2.cp39-win_amd64.pyd']},
    install_requires=['pybind11'],
    python_requires=">=3.9",
    data_files=[('Lib/site-packages/dann5', ['x64/Release/d5o2.cp39-win_amd64.pyd'])],
#    scripts=['dann5/d5o2.cp39-win_amd64.pyd']                  # The name of your scipt, and also the command you'll be using for calling it
)


# -*- coding: utf-8 -*-
"""
Created on Fri Oct 15 13:16:43 2021

@author: Nebojsa.Vojinovic
"""
# Take an array of container weights and return a Problem object that represents the cost function
from typing import List
from azure.quantum.optimization import Problem, ProblemType, Term

# Submit problem to Azure Quantum using the ParallelTempering solver:
from azure.quantum.optimization import ParallelTempering

from dann5.azureWS import workspace

class QuantumRequest:
  def __init__(self, assignmnet, timeout=100):
    self.assign = assignmnet
    self.tm = timeout
    self.problem = None
    self.sample = None
    self.nodes = None
    self.terms = None

  def reset(self):
    self.problem = None
    self.sample = None
    self.nodes = None
    self.terms = None
    self.assignment.reset()
    
  def execute(self):
    qubo = self.assign.qubo()
    self.qubo2problem(qubo)
    
    # Instantiate a solver to solve the problem.
    solver = ParallelTempering(workspace, timeout = self.tm)
    
    # Optimize the problem
    result = solver.optimize(self.problem)
#    print(f'\nResult:\n{result}\n')
    self.sample = result['configuration']
    self.solve()
    #    print_result_summary(result, nodes)
    
  def qubo2problem(self, qubo) -> Problem:
    self.terms: List[Term] = []
    self.nodes = []
    
    for (node0, node1), energy in qubo.items():
        if node0 == node1:
            self.nodes.append(node0)
#    print(self.nodes)
    
    # Expand the squared summation
    for i in range(len(self.nodes)):
        for j in range(len(self.nodes)):
            energy = qubo.get((self.nodes[i], self.nodes[j]))
#            print(i, ".", self.nodes[i], j, ".", self.nodes[j], "=", energy)
            if energy != None:
                self.terms.append(
                    Term(
                        c = energy,
                        indices = [i, j]
                    )
                )
#    print(self.terms)
    # Return an Ising-type problem
    self.problem = Problem(name="Qubo conversion", 
                           problem_type=ProblemType.pubo, 
                           terms=self.terms)
    return self.problem
    
        
  def solve(self):
    # Print a summary of the result
    d5osample = {}
    for qbit in self.sample:
        value = self.sample[qbit]
        name = self.nodes[int(qbit)]
        d5osample[name] = value
    samples = []
    samples.append(d5osample)
 #   print(f'The samples are: {samples}')
    self.assign.add(samples)
